#include "softmodem.h"
#define PUT(a) fprintf(stderr, a)
#define PUT2(a, b) fprintf(stderr, a, b)
#define OUTPUT_16BIT

// based on fsk-gen.htm

#define PI	3.141592
#define PI2	2 * PI


//Sint16 *fullBuffer;
//Sint8 *fullBuffer;
void *fullBuffer;
long fullBufferOfs = 0;
long fullBufferLen = 0;

void *softmodem_renderFrq(void*buffer, int freq, int samples) {
	//PUT2("softmodem_render_frq(): frq=%d\n", freq);
	
#ifdef OUTPUT_16BIT
	// 16 Bit
	Sint16 *s = (Sint16 *)buffer;
	const int amplitude = 256*128 - 1;		// Depends on the sample data type, e.g. 127 for Uint8
#else
	// 8 Bit
	Sint8 *s = buffer;
	const int amplitude = 127;		// Depends on the sample data type, e.g. 127 for Uint8
#endif
	
	//Uint8 smp, oldSmp;
	int smp;
	float f = PI2 * (float)freq / (float)SOFTMODEM_SAMPLERATE;
	for (int i = 0; i < samples; i++) {
		//fprintf(stderr, "Synthia::render(): a=%i / %i\n", a, len);
		
		//smp = sin(PI2 * (float)i / (float)interval) * 0.5f * amplitude;
		smp = sin(f*(float)i) * amplitude;
		
		*s = smp;
		s++;	// works for 8 and 16 Bit
	}
	return s;
}



/*
	Renders the given data as FSK into the local buffer
*/
void softmodem_queueData(char *data, int len) {
	PUT("softmodem_queueData()...\n");
	//if (str.length == 0) return;
	//var utf8 = toUTF8(str);
	
	long size = (SOFTMODEM_CARRIER_PRE + SOFTMODEM_BITS_PER_SYMBOL*len + SOFTMODEM_CARRIER_POST) * SOFTMODEM_SPB;
	//var data = "RIFF" + chr32(size+36) + "WAVE" + "fmt " + chr32(16, 0x00010001, sampleRate, sampleRate, 0x00080001) + "data" + chr32(size);
	
	// Create full wave
	fullBufferLen = 0;
	fullBufferOfs = 0;
	void *bufferCurrent = fullBuffer;
	
	
	// Pre-Carrier...
	bufferCurrent = softmodem_renderFrq(bufferCurrent, SOFTMODEM_FRQ_HI, SOFTMODEM_CARRIER_PRE*SOFTMODEM_SPB);
	
	// Data...
	char *dataCurrent = data;
	for (int j = 0; j < len; j++) {
		int c = (*dataCurrent << 1) | 0x200;	// Push Bit (?)
		for (int i = 0; i < SOFTMODEM_BITS_PER_SYMBOL; i++) {
			bufferCurrent = softmodem_renderFrq(bufferCurrent, (c % 2 == 1) ? SOFTMODEM_FRQ_HI : SOFTMODEM_FRQ_LO, SOFTMODEM_SPB);
			//c = c >>> 1;
			c = c >> 1;
		}
		dataCurrent++;
	}
	
	// Post-Carrier...
	bufferCurrent = softmodem_renderFrq(bufferCurrent, SOFTMODEM_FRQ_HI, SOFTMODEM_CARRIER_POST*SOFTMODEM_SPB);
	//if (size+44 != data.length) alert("wrong size: " + size+44 + " != " + data.length);
	
	
	fullBufferLen += size;
	PUT2("Queued %d samples\n", fullBufferLen);
}



void softmodem_start() {
#ifdef OUTPUT_16BIT
	fullBuffer = malloc(SOFTMODEM_FULLBUFFER_SAMPLES * 2);
#else
	fullBuffer = malloc(SOFTMODEM_FULLBUFFER_SAMPLES);
#endif
	if (!fullBuffer) {
		PUT("Error creating buffer. Fatal.\n");
		//fprintf(stderr, "Error creating buffer. Fatal.");
	} else {
		PUT2("Reserved %i samples of audio buffer.\n", SOFTMODEM_FULLBUFFER_SAMPLES);
	}
}
void softmodem_stop() {
	PUT("softmodem_stop()...");
	
	if (fullBuffer) free(fullBuffer);
	
	PUT("OK\n");
}




/*
	Called by the audio thread to fill the next audio buffer with data
*/
void softmodem_fillBuffer(void *buffer, int samples) {
	//PUT("softmodem_fillBuffer\n");
#ifdef OUTPUT_16BIT
	// 16 Bit
	Sint16 *s = (Sint16 *)buffer;
	Sint16 *fullBufferP = (Sint16 *)fullBuffer;
	//const int amplitude = 256*128 - 1;		// Depends on the sample data type, e.g. 127 for Uint8
#else
	// 8 Bit
	Sint8 *s = (Sint8 *)buffer;
	Sint8 *fullBufferP = (Sint8 *)fullBuffer;
	//const int amplitude = 127;		// Depends on the sample data type, e.g. 127 for Uint8
#endif
	
	fullBufferP += fullBufferOfs;
	
	for (int i = 0; i < samples; i++) {
		
		if (fullBufferOfs < fullBufferLen) {
			*s = *fullBufferP;
			fullBufferP++;
			fullBufferOfs++;
		} else {
			*s = 0;
		}
		
		s++;	// works for 8 and 16 Bit
	}
}

int softmodem_isActive() {
	return (fullBufferOfs < fullBufferLen);
}

