/**
 * Softmodem Test
 * 2013 by Bernhard Slawik Digital, http://www.bernhardslawik.de
 *
 **/

#define PUT	putError
#define PUTERROR putError
#define OUTPUT_16BIT
#define AUDIO_BUFFER 512


// General Includes
#include <stdio.h>
#include <math.h>
#include <sstream>
using namespace std;


#include "SDL/SDL_ttf.h"
#include "SDL.h"
#include "PDL.h"
#include "softmodem.h"


// Network
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <arpa/inet.h>



// SDL_vnc includes
#ifdef WIN32
	#include <windows.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>



SDL_Surface *Surface;               // Screen surface to retrieve width/height information
TTF_Font *font;
SDL_Surface *textMap;
SDL_Rect textMapRect;


//char *text = strdup("Hello world! This is softmodemTest!");
int hintTime = 0;
int hint = 0;
const int hints = 6;
char *hintText[hints] = {
	strdup("SoftGolem"),
	strdup("FSK Softmodem for webOS!"),
	strdup("This phone connects to TCP..."),
	strdup("...and relays traffic to audio"),
	strdup("(C) 2013 Bernhard Slawik Digital"),
	strdup("http://www.bernhardslawik.de")
};

//bool invalid;
const int fps = 60;


// My army of little helpers...

int toInt(string const & s) {
	stringstream ss;
	int val;
	ss << s;
	ss >> val;
	return val;
}

string toStr(int i) {
	stringstream ss;
	string str;
	ss << i;
	getline(ss, str);
	return str;
}

float sqr(float r) {
	return r*r;
}
float clip(float rmin, float rmax, float r) {
	if (r < rmin) return rmin;
	else if (r > rmax) return rmax;
	else return r;
}

// Debug functionality
void putError(const char * format, ...) {
	char s[512];
	va_list args;
	va_start(args, format);
	vsprintf(s, format, args);
	//fprintf(stderr, ">>> %s", s);
	fprintf(stderr, "%s", s);
	va_end(args);
}


////////////////////////////////////////////////////////////////////////////////
// Plugin
////////////////////////////////////////////////////////////////////////////////

volatile float RotationSpeed = 0.001f; // Rotation speed of our object

// another cross-thread variable, used to stop movement when plugin is in card that's minimized
volatile bool Paused = false;

/*
Call a JS function:
	const char *params[2];
	params[0] = "foo";
	params[1] = "bar";
	PDL_Err mjErr = PDL_CallJS("testFunc", params, 2);
	if ( mjErr != PDL_NOERROR )
	{
		printf("error\n");
	}
*/

PDL_bool sendData(PDL_JSParameters *params) {
	if (PDL_GetNumJSParams(params) < 1) {
		PDL_JSException(params, "You must send one string value");
		return PDL_TRUE;
	}
	
	char *val = strdup(PDL_GetJSParamString(params, 0));
	
	char reply[32];
	snprintf(reply, 32, "Sending %i bytes...", strlen(val));
	PDL_JSReply(params, reply);
	
	softmodem_queueData(&val[0], strlen(val));
	
	return PDL_TRUE;
}


void doSetRotationSpeed(double val) {
	// we'll make the slowest it goes 0.001, and 
	// the fastest 0.01
	double slowest = -0.005;
	double fastest = 0.005;
	double range = fastest - slowest;

	RotationSpeed = (float)(slowest + ((range*val)/100.0f));
	
	char buf[2];
	buf[0] = range;
	buf[1] = 255 * val/100.0f;
	softmodem_queueData(&buf[0], 1);
}

// takes a value from 0 to 100 and make it 0 to 2*PI
PDL_bool setRotationSpeed(PDL_JSParameters *params) {
	double val;
	if (PDL_GetNumJSParams(params) < 1 ||
		(val = PDL_GetJSParamDouble(params, 0), val < 0.0 || val > 100.0))
	{
		PDL_JSException(params, "You must send a value from 0 to 100");
		return PDL_TRUE;
	}
	
	char reply[32];
	snprintf(reply, 32, "%2.3f", val);
	PDL_JSReply(params, reply);
	
	doSetRotationSpeed(val);

	return PDL_TRUE;
}


// these methods take no parameters.  They're needed when running as a plugin in a hybrid
// application to pause the rendering loop because plugins to JS apps don't get application
// activation messages.
PDL_bool pause(PDL_JSParameters *params) {
	Paused = true;
}

PDL_bool resume(PDL_JSParameters *params) {
	Paused = false;

	// push a bogus event into the queue to wake up main thread
	SDL_Event Event;
	Event.active.type = SDL_ACTIVEEVENT;
	Event.active.gain = 1;
	Event.active.state = 0; /* no state makes this event meaningless */
	SDL_PushEvent(&Event);
}



////////////////////////////////////////
// Server
////////////////////////////////////////
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 

#define MY_PORT		9999
#define MAXBUF		1024
int sockfd;

void ServerStart() {
	PUT("ServerStart()...\n");
	
	/*
	// Make it a server (listening server)
	
	int sockfd;
	struct sockaddr_in self;
	char buffer[MAXBUF];
	
	
	// Create streaming socket
	if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
		PUTERROR("Socket");
		return;
	}

	// Initialize address/port structure
	bzero(&self, sizeof(self));
	self.sin_family = AF_INET;
	self.sin_addr.s_addr = INADDR_ANY;
	self.sin_port = htons(MY_PORT);
	
	// Assign a port number to the socket
	if (bind(sockfd, (struct sockaddr*)&self, sizeof(self)) != 0 ) {
		PUTERROR("socket--bind");
		return;
	}
	
	
	if (listen(sockfd, 20) != 0 ) {
	//if (listen(sockfd, 10) != 0 ) {
	//if (listen(sockfd, 5) != 0 ) {
		PUTERROR("socket--listen");
		return;
	}

	// Forever... 
	//while (1) {
		int clientfd;
		struct sockaddr_in client_addr;
		int addrlen = sizeof(client_addr);
		
		PUT("Waiting for client on port %d...", MY_PORT);
		// accept a connection (creating a data pipe)
		clientfd = accept(sockfd, (struct sockaddr*)&client_addr, (socklen_t*)&addrlen);
		PUT("%s:%d connected\n", inet_ntoa(client_addr.sin_addr), ntohs(client_addr.sin_port));
		
		// Echo back anything sent
		send(clientfd, buffer, recv(clientfd, buffer, MAXBUF, 0), 0);
		
		// Close data connection
		close(clientfd);
	//}
	
	PUT("closing socket...");
	// Clean up (should never get here!)
	close(sockfd);
	*/
	
	
	// Make a client connection (e.g. because firewall blocks incoming connections)
	//int sockfd, portno, n;
	
	struct sockaddr_in serv_addr;
	struct hostent *server;
	
	char *host = strdup("192.168.4.15");
	int portno = 9998;
	
	int n = 0;
	
	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		PUTERROR("ERROR opening socket");
		return;
	}
	server = gethostbyname(host);
	if (server == NULL) {
		PUTERROR("ERROR, no such host\n");
		return;
	}
	
	PUT("Un-blocking socket...");
	// non-blocking
	fcntl(sockfd, F_SETFL, O_NONBLOCK);
	
	PUT("Connecting to %s:%d...", host, portno);
	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	bcopy((char *)server->h_addr,	(char *)&serv_addr.sin_addr.s_addr,	server->h_length);
	serv_addr.sin_port = htons(portno);
	n = connect(sockfd,	(struct sockaddr *)&serv_addr, sizeof(serv_addr));
	if (n < 0) {
		PUTERROR("ERROR connecting");
		return;
	}
	PUT("OK\n");
	
	/*
	PUT("Please enter the message: ");
	bzero(buffer, 256);
	fgets(buffer, 255, stdin);
	
	n = write(sockfd, buffer, strlen(buffer));
	if (n < 0) PUTERROR("ERROR writing to socket");
	*/
	
	/*
	PUT("Waiting for data from other side...\n");
	bzero(buffer, 256);
	n = read(sockfd, buffer, 255);
	if (n < 0) PUTERROR("ERROR reading from socket");
	PUT("%s\n",buffer);
	*/
	
	
	PUT("OK\n");
}

void ServerUpdate() {
	char buffer[256];
	
	//  Wait for a connection, then accept() it
	int flags = 0;
	int n = recv(sockfd, buffer, sizeof(buffer), flags);
	if (n > 0) {
		PUT("ServerUpdate(): Received %d bytes: ", n);
		
		// do something...
		buffer[n] = 0;
		
		PUT("[");
		PUT(strdup(&buffer[0]));
		PUT("]\n");
		
		softmodem_queueData(strdup(&buffer[0]), n);
		
	} else {
		//PUT("No data received...\n");
	}
}

void ServerStop() {
	if (sockfd) {
		close(sockfd);
		sockfd = 0;
	}
}

////////////////////////////////////////
// Audio
////////////////////////////////////////


const int visBufferLen = 320;
int visBufferOfs = 0;
int visBufferOfsRead = 0;
Sint8 visBuffer[visBufferLen];


//long testi = 0;
void PlayCallback(void *ud, Uint8 *stream, int len) {
	SDL_LockAudio();
	int i;
	
#ifdef OUTPUT_16BIT
	len = len >> 1;	// 16 Bit
	Sint16 *s = (Sint16 *)stream;
#else
	Sint8 *s = (Sint8 *)stream;
#endif
	
	//softmodem_render((Sint8 *)stream, len);
	softmodem_fillBuffer(stream, len);
	
	
	// Save wave form for visualization
	visBufferOfsRead = visBufferOfs;
	for (i = 0; i < len; i++) {
		
		// Test 16-Bit beep
		//*s = sin(testi * 0.12f) * 256*40;
		//testi++;
		
#ifdef OUTPUT_16BIT
		// 16 Bit
		visBuffer[visBufferOfs] = (*s) >> 8;
#else
		// 8 Bit
		visBuffer[visBufferOfs] = *s;
#endif
		visBufferOfs = (visBufferOfs + 1) % visBufferLen;
		s++;
	}
	
	SDL_UnlockAudio();
}

void StartAudio() {
	PUT("AudioSpec...\n");
	SDL_AudioSpec specDesired, *specObtained;
	specObtained = (SDL_AudioSpec*)malloc(sizeof(SDL_AudioSpec));
	//memset(specDesired, 0, sizeof(SDL_AudioSpec));
	//specDesired.freq = 44100;
	specDesired.freq = SOFTMODEM_SAMPLERATE;
#ifdef OUTPUT_16BIT
	specDesired.format = AUDIO_S16;
#else
	specDesired.format = AUDIO_S8;	// Signed 8 Bit
#endif
	specDesired.channels = 1;
	specDesired.samples = AUDIO_BUFFER;	//512;	//1024; //4096;
	specDesired.userdata = NULL;
	specDesired.callback = PlayCallback;
	
	PUT("AudioSpecs (desired): %i Hz, %i BitMode, %i Channels, %i Samples\n", specDesired.freq, specDesired.format, specDesired.channels, specDesired.samples);
	PUT("OpenAudio...\n");
	if (SDL_OpenAudio(&specDesired, specObtained) != 0) {
	//if (SDL_OpenAudio(&specDesired, NULL) != 0) {
		PUTERROR("SDL_OpenAudio failed: %s\n", SDL_GetError());
		return;
	}
	
	PUT("AudioSpecs (obtained): %i Hz, %i BitMode, %i Channels, %i Samples\n", specObtained->freq, specObtained->format, specObtained->channels, specObtained->samples);
	free(specObtained);
	
	// Play
	PUT("Play...\n");
	SDL_PauseAudio(0);
}

void StopAudio() {
	SDL_CloseAudio();
}



////////////////////////////////////////
// Graphics
////////////////////////////////////////

Uint32 GetColor(SDL_Surface *screen, Uint32 color) {
	return SDL_MapRGB(screen->format, (color & 0xff0000) >> 16, (color & 0x00ff00) >> 8, (color & 0x0000ff));
}

void SetPixel(SDL_Surface *screen, int x, int y, Uint32 color) {
	switch (screen->format->BytesPerPixel) {
		case 1: // Assuming 8-bpp
			Uint8 *bufp8;
			bufp8 = (Uint8 *)screen->pixels + y*screen->pitch + x;
			*bufp8 = color;
			break;
		case 2: // Probably 15-bpp or 16-bpp
			Uint16 *bufp16;
			bufp16 = (Uint16 *)screen->pixels + y*screen->pitch/2 + x;
			*bufp16 = color;
			break;
		case 3: // Slow 24-bpp mode, usually not used
			Uint8 *bufp3;
			bufp3 = (Uint8 *)screen->pixels + y*screen->pitch + x * 3;
			if(SDL_BYTEORDER == SDL_LIL_ENDIAN)
			{
				bufp3[0] = color;
				bufp3[1] = color >> 8;
				bufp3[2] = color >> 16;
			} else {
				bufp3[2] = color;
				bufp3[1] = color >> 8;
				bufp3[0] = color >> 16;
			}
			break;
		case 4: // Probably 32-bpp
			Uint32 *bufp4;
			bufp4 = (Uint32 *)screen->pixels + y*screen->pitch/4 + x;
			*bufp4 = color;
			break;
	}
}

void CreateText(SDL_Surface **textSurface, char *s, int cr, int cg, int cb) {
	if (*textSurface != NULL) SDL_FreeSurface(*textSurface);
	SDL_Color color = {cr, cg, cb};
	*textSurface = TTF_RenderText_Solid(font, s, color);
	
	/*
	textSurfaceRect->x = x; //(Surface->w - text->w) / 2;
	textSurfaceRect->y = y; //(Surface->h - text->h) / 2;
	textSurfaceRect->w = *textSurface->w;
	textSurfaceRect->h = *textSurface->h;
	*/
}


void Redraw() {
	//if (!invalid) return;
	
	unsigned int color;
	// SDL Redraw
	
	// Show mode
	//SDL_FillRect(Surface, NULL, GetColor(Surface, 0x33cc99));
	if (softmodem_isActive()) {
		SDL_FillRect(Surface, NULL, GetColor(Surface, 0xcc9933));
		// Draw wave form
		color = GetColor(Surface, 0xffffff);
		int x;
		int i;
		int const amplitude = 127;
		float ys = (float)Surface->h / (amplitude * 2.0f);	// depends on audio mode (8 / 16 bit signed/unsigned)
		//float xs = (float)Surface->w / visBufferLen;
		//for (i = 0; i < visBufferLen; i++) {
		for (x = 0; x < Surface->w; x++) {
			int y = (visBuffer[(visBufferOfsRead + x) % visBufferLen] + amplitude) * ys;
			SetPixel(Surface, x, y, color);
		}
	} else {
		SDL_FillRect(Surface, NULL, GetColor(Surface, 0x3399cc));
	}
	
	// Show text
	textMapRect.x = 8;
	textMapRect.y = 8;
	textMapRect.w = textMap->w;
	textMapRect.h = textMap->h;
	SDL_BlitSurface(textMap, NULL, Surface, &textMapRect);
	
	// Update
	SDL_Flip(Surface);
}




////////////////////////////////////////
// Main GUI loop
////////////////////////////////////////
#define MAX_CMD_BUF 256
char cmdBuf[MAX_CMD_BUF+1];
int cmdBufO = 0;



void Run(SDL_Surface *screen) {
	PUT("Run()...\n");
	
	SDL_Event event;
	int inloop;
	unsigned int color;
	int mx, mtx, mtxOld, mtxDown;
	int my, mty, mtyOld, mtyDown;
	int mtb = 0;
	int mtn = 0;
	int mtnOld = 0;
	
	float rX, rY;
	float dX, dY;
	
	inloop = 1;
	
	PUT("Run(): Entering Loop...\n");
	Redraw();
	
	while (inloop) {
		
		// Check for events
		
		while (SDL_PollEvent(&event)) {
			
			switch (event.type) {
				case SDL_KEYDOWN:
					switch(event.key.keysym.sym) {
						//case PDLK_GESTURE_FORWARD: break;
						//case PDLK_GESTURE_BACK: break;
						//case PDLK_GESTURE_AREA: break;
						
						case '@':
							softmodem_queueData(strdup("The quick brown fox jumps over the lazy dog"), 43);
							break;
						
						case 8:
							if (cmdBufO > 0) cmdBufO--;
							
							if (cmdBufO > 0) {
								cmdBuf[cmdBufO] = 0;
								CreateText(&textMap, strdup(&cmdBuf[0]), 0xff, 0xff, 0x00);
							} else {
								CreateText(&textMap, strdup(" "), 0xff, 0xff, 0x00);
							}
							Redraw();
							hintTime = 0;
							break;
						
						case 10:
						case 13:
							// Append LF
							cmdBuf[cmdBufO++] = 0x0a;
						
							softmodem_queueData(&cmdBuf[0], cmdBufO);
							cmdBufO = 0;
							CreateText(&textMap, strdup("OK!"), 0x00, 0xff, 0x00);
							Redraw();
							hintTime = 0;
							break;
						
						case 27:	// Gesture: back
							break;
						case 229:	// Gesture: forward
							break;
						case 231:	// Gesture-Area
							break;
						case 304:	// Shift
							break;
						case 305:	// Sym-key
							break;
						case 307:	// Orange-key
							break;
						
						default:
							PUT("key %d\n", event.key.keysym.sym);
							if (cmdBufO < MAX_CMD_BUF) {
								cmdBuf[cmdBufO] = event.key.keysym.sym;
								cmdBufO++;
							}
							
							cmdBuf[cmdBufO] = 0;
							CreateText(&textMap, strdup(&cmdBuf[0]), 0xff, 0xff, 0xff);
							Redraw();
							hintTime = 0;
							break;
						
						//case 'q':	inloop = false;	break;
						
						/*
						case 8: break;
						*/
					}
					break;
					
				case SDL_KEYUP:
					//if (event.key.keysym.sym == (SDLKey)PDLK_GESTURE_AREA) {}
					break;
					
					
				// Test multi touch stuff
				case SDL_MOUSEBUTTONDOWN:
				case SDL_MOUSEBUTTONUP:
				case SDL_MOUSEMOTION:
					mtn = 0;
					mtx = 0;
					mty = 0;
					int mxs[4];
					int mys[4];
					while ((mtn < 4) && (SDL_GetMultiMouseState(mtn, &mx, &my) != 0)) {
						mxs[mtn] = mx;
						mys[mtn] = my;
						mtx += mx;
						mty += my;
						mtn ++;
					}
					
					if (mtn > 0) {
						mtx /= mtn;
						mty /= mtn;
						switch (mtn) {
							case 1:	// Single finger
								if (mtnOld == 0) {
									// MouseDown
									mtxDown = mtx;
									mtyDown = mty;
									mtxOld = mtx;
									mtyOld = mty;
								}
								
								dX = (float)(mtx - mtxOld) / (float)screen->w;
								dY = (float)(mty - mtyOld) / (float)screen->h;
								rX = (float)mtx / (float)screen->w;
								rY = (float)mty / (float)screen->h;
								//PUT("mtx=%i, sw=%i, rX=%f -- mty=%i, sh=%i, rY=%f\n", mtx, screen->w, rX,  mty, screen->h, rY);
								
								if ((event.type == SDL_MOUSEMOTION) && (mtnOld == 1)) {
									// Do something with dX, dY
								}
								mtxOld = mtx;
								mtyOld = mty;
								break;
							
							case 2:	// 2-Finger
								//if (event.type == SDL_MOUSEBUTTONDOWN) {
									if (mtnOld < 2) {
										//ToggleMode((mxs[1] > mxs[0]) ? 1 : -1 );
									}
								//}
								break;
							
						}
					}
					mtnOld = mtn;
					break;
				
				case SDL_ACTIVEEVENT:
					PUT("SDL_ACTIVEEVENT\n"); //: gain=%i, state=%i\n", event.gain, event.state);
					if (SDL_GetAppState() & SDL_APPACTIVE) {
						SDL_PauseAudio(0);
						CreateText(&textMap, strdup("Welcome back!"), 0xff, 0xff, 0xff);
						//Invalidate();
						Redraw();
					} else {
						SDL_PauseAudio(1);
						CreateText(&textMap, strdup("PAUSE - PAUSE - PAUSE - PAUSE"), 0x00, 0x00, 0x00);
						//Invalidate();
						Redraw();
					}
					break;
				case SDL_VIDEOEXPOSE:
					Redraw();
					break;
				
				case SDL_QUIT:
					PUT("SDL_QUIT received, exiting loop...\n");
					inloop = 0;
					break;
			}
		}
		
		
		ServerUpdate();
		
		if ((SDL_GetAppState() & SDL_APPACTIVE) && (!Paused)) {
			// Active
			hintTime++;
			if (hintTime > fps * 5) {
				hintTime = 0;
				hint = (hint + 1) % hints;
				CreateText(&textMap, hintText[hint], 0x00, 0x00, 0x00);
			}
			
			SDL_Delay(1000 / fps);
			Redraw();
			
		} else {
			// Idle
			SDL_Delay(100);
			
		}
	}
	PUT("Run(): Stopped...\n");
}



////////////////////////////////////////////////////////////////////////////////
// Main
////////////////////////////////////////////////////////////////////////////////


int main(int argc, char** argv) {
	// Title
	PUT("%s by Bernhard Slawik Digital, http://www.bernhardslawik.de\n", hintText[0]);
	
	//@FIXME: Obtain screen dimensions from PDK!
	int startGUI = 1;
	char *startData = NULL;
	
	PUT("Softmodem...\n");
	softmodem_start();
	/*
	if (argc > 1) {
		startData = argv[1];
		if (strlen(startData) > 0) {
			PUT("Command line arguments: [");
			PUT(startData);
			PUT("]\n");
			
			startGUI = 0;
		}
	}
	*/
	
	
	
	int w = 320;
	int h = 80;	//480;
	int desired_bpp = 0;
	Uint32 video_flags = 0;

	// Force double buffering
	//video_flags |= SDL_DOUBLEBUF;
	
	// webOS init:
	int result;
	PUT("SDL_Init...\n");
	// Initialize the SDL library with the Video subsystem
	//result = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_NOPARACHUTE);
	result = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO);
	
	if (result != 0) {
		PUT("SDL_Init Error: %s\n", SDL_GetError());
		exit(1);
	}

	// start the PDL library
	PUT("PDL_Init...\n");
	PDL_Init(0);
	
	
	if (startGUI == 1) {
		
		PUT("Plug-In exports...\n");
		// register the js callback
		PDL_RegisterJSHandler("sendData", sendData);
		PDL_RegisterJSHandler("setRotationSpeed", setRotationSpeed);
		PDL_RegisterJSHandler("setAngle", setRotationSpeed);
		PDL_RegisterJSHandler("pause", pause);
		PDL_RegisterJSHandler("resume", resume);
		PDL_JSRegistrationComplete();
		
		
		PUT("Server...\n");
		ServerStart();
		
		// GUI mode
		PUT("Loading TTF support...\n");
		// Init TTF support
		if (TTF_Init() == -1) {
			PUTERROR("Unable to initialize SDL_ttf: %s \n", TTF_GetError());
			exit(1);
		}
		PUT("Loading font...\n");
		SDL_Color color2 = {0, 0, 0};
		font = TTF_OpenFont("04b_09.ttf", 12);
		
		// SDL
		PUT("Initializing SDL GUI...\n");
		//Surface = SDL_SetVideoMode(w, h, desired_bpp, video_flags);
		Surface = SDL_SetVideoMode(0, 0, 0, video_flags);
		if (Surface == NULL) {
			PUT("Error creating surface for SDL: %s\n", SDL_GetError());
			PDL_Quit();
			SDL_Quit();
			exit(1);
		}
		
		// Show Splash text
		//CreateText(textMap, strdup("HTK303 - Welcome!"), 0xff, 0xff, 0xff);
		CreateText(&textMap, hintText[0], 0xff, 0xff, 0xff);
		textMapRect.x = 8;
		textMapRect.y = 8;
		textMapRect.w = textMap->w;
		textMapRect.h = textMap->h;
		SDL_BlitSurface(textMap, NULL, Surface, &textMapRect);
		SDL_Flip(Surface);
	}
	
	
	PUT("Starting audio...\n");
	StartAudio();
	
	// Ready!
	if (startData != NULL) {
		// Queue initial data
		softmodem_queueData(startData, strlen(startData));
	}
	
	if (startGUI == 1)
		Run(Surface);
	else
		SDL_Delay(
			1000 * SOFTMODEM_SPB *
			(
				SOFTMODEM_CARRIER_PRE + strlen(startData) * SOFTMODEM_BITS_PER_SYMBOL + SOFTMODEM_CARRIER_POST
			)
			/ SOFTMODEM_SAMPLERATE
			
			+ 100
		);
	
	
	// Cleanup
	PUT("Stopping...\n");
	
	StopAudio();
	SDL_Delay(1000*AUDIO_BUFFER/SOFTMODEM_SAMPLERATE);
	
	softmodem_stop();
	
	
	PUT("GUI quit...");
	if (startGUI == 1) {
		ServerStop();
		
		PUT("TTF...");
		TTF_Quit();
	}
	
	PUT("PDL...");
	PDL_Quit();
	
	PUT("SDL...");
	SDL_Quit();
	
	PUT("OK\nGood bye.\n");
	return 0;
}
